"""Module for PacBio fixtures returning Path objects."""

from pathlib import Path

import pytest

from cg.constants.pacbio import PacBioDirsAndFiles
from cg.services.run_devices.pacbio.run_file_manager.models import PacBioRunValidatorFiles


# Directory fixtures
@pytest.fixture
def pac_bio_fixtures_dir(devices_dir: Path) -> Path:
    """Return the path to the PacBio fixtures directory."""
    return Path(devices_dir, "pacbio")


@pytest.fixture
def pac_bio_wrong_metrics_dir(pac_bio_fixtures_dir: Path) -> Path:
    """Return the path to the PacBio metrics directory."""
    return Path(pac_bio_fixtures_dir, "wrong_metrics")


@pytest.fixture
def pac_bio_runs_dir(pac_bio_fixtures_dir: Path) -> Path:
    """Return the path to the PacBio run directory."""
    return Path(pac_bio_fixtures_dir, "SMRTcells")


@pytest.fixture
def pac_bio_test_run_dir(pac_bio_runs_dir: Path, pac_bio_test_run_name: str) -> Path:
    """Return the path to a PacBio run directory."""
    return Path(pac_bio_runs_dir, pac_bio_test_run_name)


@pytest.fixture
def pacbio_barcoded_run_dir(pac_bio_runs_dir: Path, pacbio_barcoded_run_name: str) -> Path:
    """Return the path to a PacBio run directory."""
    return Path(pac_bio_runs_dir, pacbio_barcoded_run_name)


@pytest.fixture
def pac_bio_smrt_cell_dir_1_b01(pac_bio_test_run_dir: Path, pac_bio_smrt_cell_name: str) -> Path:
    """Return the path to a PacBio SMRT cell directory."""
    return Path(pac_bio_test_run_dir, pac_bio_smrt_cell_name)


@pytest.fixture
def pacbio_barcoded_smrt_cell_dir_1_c01(
    pacbio_barcoded_run_dir: Path, pacbio_barcoded_smrt_cell_name: str
) -> Path:
    """Return the path to a barcoded PacBio SMRT cell directory."""
    return Path(pacbio_barcoded_run_dir, pacbio_barcoded_smrt_cell_name)


@pytest.fixture
def pacbio_barcoded_hifi_reads_dir(pacbio_barcoded_smrt_cell_dir_1_c01: Path) -> Path:
    """Return the path to a PacBio HiFi reads directory."""
    return Path(pacbio_barcoded_smrt_cell_dir_1_c01, PacBioDirsAndFiles.HIFI_READS)


@pytest.fixture
def pac_bio_run_statistics_dir_1_b01(pac_bio_smrt_cell_dir_1_b01: Path) -> Path:
    """Return the path to the PacBio SMRT cell statistics directory."""
    return Path(pac_bio_smrt_cell_dir_1_b01, PacBioDirsAndFiles.STATISTICS_DIR)


@pytest.fixture
def pacbio_barcoded_run_statistics_dir(pacbio_barcoded_smrt_cell_dir_1_c01: Path) -> Path:
    """Return the path to the statistics directory of the PacBio barcoded SMRT cell."""
    return Path(pacbio_barcoded_smrt_cell_dir_1_c01, PacBioDirsAndFiles.STATISTICS_DIR)


@pytest.fixture
def pac_bio_run_reports_dir_1_b01(pac_bio_run_statistics_dir_1_b01: Path) -> Path:
    """Return the path to the PacBio SMRT cell unzipped_reports directory"""
    return Path(pac_bio_run_statistics_dir_1_b01, PacBioDirsAndFiles.UNZIPPED_REPORTS_DIR)


@pytest.fixture
def pacbio_barcoded_run_reports_dir(pacbio_barcoded_run_statistics_dir: Path) -> Path:
    """Return the path to the unzipped reports directory of the PacBio barcoded SMRT cell."""
    return Path(pacbio_barcoded_run_statistics_dir, PacBioDirsAndFiles.UNZIPPED_REPORTS_DIR)


@pytest.fixture
def pac_bio_run_metadata_dir_1_b01(pac_bio_smrt_cell_dir_1_b01: Path) -> Path:
    """Return the path to the PacBio SMRT cell metadata directory"""
    return Path(pac_bio_smrt_cell_dir_1_b01, PacBioDirsAndFiles.METADATA_DIR)


# File fixtures


@pytest.fixture
def pacbio_barcodes_report_file(
    pacbio_barcoded_run_reports_dir: Path,
) -> Path:
    """Return the path to the PacBio barcodes report file."""
    return Path(pacbio_barcoded_run_reports_dir, PacBioDirsAndFiles.BARCODES_REPORT)


@pytest.fixture
def pacbio_barcoded_ccs_report_file(
    pacbio_barcoded_run_statistics_dir: Path, pacbio_barcoded_1_c01_cell_full_name: str
) -> Path:
    """Return the path to the PacBio CCS report file."""
    return Path(
        pacbio_barcoded_run_statistics_dir,
        f"{pacbio_barcoded_1_c01_cell_full_name}.{PacBioDirsAndFiles.CCS_REPORT_SUFFIX}",
    )


@pytest.fixture
def pacbio_barcoded_control_report_file(pacbio_barcoded_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio loading report file."""
    return Path(pacbio_barcoded_run_reports_dir, PacBioDirsAndFiles.CONTROL_REPORT)


@pytest.fixture
def pacbio_barcoded_loading_report_file(pacbio_barcoded_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio loading report file."""
    return Path(pacbio_barcoded_run_reports_dir, PacBioDirsAndFiles.LOADING_REPORT)


@pytest.fixture
def pacbio_barcoded_raw_data_report_file(pacbio_barcoded_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio loading report file."""
    return Path(pacbio_barcoded_run_reports_dir, PacBioDirsAndFiles.RAW_DATA_REPORT)


@pytest.fixture
def pacbio_barcoded_smrtlink_datasets_report_file(pacbio_barcoded_run_reports_dir: Path) -> Path:
    """Return the path to the PacBio loading report file."""
    return Path(pacbio_barcoded_run_reports_dir, PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT)


@pytest.fixture
def pac_bio_transferdone_file_1_b01(pac_bio_run_metadata_dir_1_b01: Path) -> Path:
    """Return the path to the PacBio SMRTLink datasets report file."""
    return Path(pac_bio_run_metadata_dir_1_b01, "m84202_240522_155607_s2.transferdone")


@pytest.fixture
def pac_bio_zipped_reports_file_1_b01(pac_bio_run_statistics_dir_1_b01: Path) -> Path:
    """Return the path to the PacBio SMRTLink datasets report file."""
    return Path(pac_bio_run_statistics_dir_1_b01, "m84202_240522_155607_s2.reports.zip")


@pytest.fixture
def pacbio_barcoded_report_files_to_parse(
    pacbio_barcodes_report_file: Path,
    pacbio_barcoded_ccs_report_file: Path,
    pacbio_barcoded_control_report_file: Path,
    pacbio_barcoded_loading_report_file: Path,
    pacbio_barcoded_raw_data_report_file: Path,
    pacbio_barcoded_smrtlink_datasets_report_file: Path,
) -> list[Path]:
    """Return the list of PacBio report files to parse."""
    return [
        pacbio_barcodes_report_file,
        pacbio_barcoded_control_report_file,
        pacbio_barcoded_loading_report_file,
        pacbio_barcoded_raw_data_report_file,
        pacbio_barcoded_smrtlink_datasets_report_file,
        pacbio_barcoded_ccs_report_file,
    ]


@pytest.fixture
def pacbio_barcoded_hifi_read_file(
    pacbio_barcoded_hifi_reads_dir: Path, pacbio_barcoded_1_c01_cell_full_name: str
) -> Path:
    """Return the Pacbio bacrcoded HiFi read file."""
    return Path(
        pacbio_barcoded_hifi_reads_dir,
        f"{pacbio_barcoded_1_c01_cell_full_name}.hifi_reads.bc2004.bam",
    )


@pytest.fixture
def pacbio_unassigned_hifi_read_file(
    pacbio_barcoded_hifi_reads_dir: Path, pacbio_barcoded_1_c01_cell_full_name: str
) -> Path:
    """Return the Pacbio unassigned HiFi read file."""
    return Path(
        pacbio_barcoded_hifi_reads_dir,
        f"{pacbio_barcoded_1_c01_cell_full_name}.hifi_reads.unassigned.bam",
    )


@pytest.fixture
def pacbio_barcoded_hifi_read_files(pacbio_barcoded_hifi_read_file: Path) -> list[Path]:
    """Return the list of PacBio HiFi read files."""
    return [pacbio_barcoded_hifi_read_file]


@pytest.fixture
def expected_1_b01_run_validation_files(
    pac_bio_transferdone_file_1_b01: Path,
    pac_bio_zipped_reports_file_1_b01: Path,
    pac_bio_run_reports_dir_1_b01: Path,
) -> PacBioRunValidatorFiles:
    """Return the expected run validation files for the 1_B01 SMRT cell."""
    return PacBioRunValidatorFiles(
        manifest_file=pac_bio_transferdone_file_1_b01,
        decompression_target=pac_bio_zipped_reports_file_1_b01,
        decompression_destination=pac_bio_run_reports_dir_1_b01,
    )


# Wrong files


@pytest.fixture
def pac_bio_wrong_metrics_file(pac_bio_wrong_metrics_dir: Path) -> Path:
    """Return the path to a temporary PacBio statistics directory."""
    return Path(pac_bio_wrong_metrics_dir, "metrics.json")



from cg.constants import Workflow
from cg.store.models import Case, Customer, Sample
from cg.store.store import Store


def test_get_related_cases(
    store_with_rna_and_dna_samples_and_cases: Store,
    related_dna_sample_1: Sample,
    rna_sample_collaborators: set[Customer],
    related_dna_cases: list[Case],
):
    # GIVEN a database with a sample in several cases
    # GIVEN a list of workflows

    workflows = [
        Workflow.MIP_DNA,
        Workflow.BALSAMIC,
        Workflow.BALSAMIC_UMI,
    ]

    # WHEN getting the cases from the given sample by the given workflows and within the given collaborators
    fetched_related_dna_cases: list[Case] = (
        store_with_rna_and_dna_samples_and_cases.get_related_cases(
            sample_internal_id=related_dna_sample_1.internal_id,
            workflows=workflows,
            collaborators=rna_sample_collaborators,
        )
    )

    # THEN the correct set of cases is returned
    assert set(related_dna_cases) == set(fetched_related_dna_cases)



"""Module for the NovaSeq6000 sequencing times service."""

from datetime import datetime
from pathlib import Path

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)
from cg.utils.files import get_source_modified_time_stamp
from cg.utils.time import format_time_from_ctime


class Novaseq6000SequencingTimesService(SequencingTimesService):
    """Class to get the sequencing times for NovaSeq6000 sequencing runs."""

    @staticmethod
    def get_start_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer start date and time."""
        return run_directory_data.sequenced_at

    @staticmethod
    def get_end_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer end date and time."""
        file_path: Path = run_directory_data.get_sequencing_completed_path
        modified_time: float = get_source_modified_time_stamp(file_path)
        return format_time_from_ctime(modified_time)



"""CLI support to create config and/or start TAXPROFILER."""

import logging

import click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.commands import resolve_compression
from cg.cli.workflow.nf_analysis import (
    config_case,
    metrics_deliver,
    report_deliver,
    run,
    start,
    start_available,
    store,
    store_available,
    store_housekeeper,
)
from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def taxprofiler(context: click.Context) -> None:
    """nf-core/taxprofiler analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = TaxprofilerAnalysisAPI(config=context.obj)


taxprofiler.add_command(resolve_compression)
taxprofiler.add_command(config_case)
taxprofiler.add_command(run)
taxprofiler.add_command(start)
taxprofiler.add_command(start_available)
taxprofiler.add_command(metrics_deliver)
taxprofiler.add_command(report_deliver)
taxprofiler.add_command(store_housekeeper)
taxprofiler.add_command(store)
taxprofiler.add_command(store_available)



"""Module for the factory of the deliver files service."""

from typing import Type

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import DataDelivery, Workflow
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.deliver_files.deliver_files_service.deliver_files_service import (
    DeliverFilesService,
)
from cg.services.deliver_files.deliver_files_service.exc import DeliveryTypeNotSupported
from cg.services.deliver_files.file_fetcher.abstract import FetchDeliveryFilesService
from cg.services.deliver_files.file_fetcher.analysis_raw_data_service import (
    RawDataAndAnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.analysis_service import AnalysisDeliveryFileFetcher
from cg.services.deliver_files.file_fetcher.raw_data_service import RawDataDeliveryFileFetcher
from cg.services.deliver_files.file_filter.sample_service import SampleFileFilter
from cg.services.deliver_files.file_formatter.abstract import DeliveryFileFormattingService
from cg.services.deliver_files.file_formatter.service import DeliveryFileFormatter
from cg.services.deliver_files.file_formatter.utils.case_service import CaseFileFormatter
from cg.services.deliver_files.file_formatter.utils.sample_concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.utils.sample_service import (
    SampleFileFormatter,
    FileManagingService,
    SampleFileNameFormatter,
)
from cg.services.deliver_files.file_mover.service import DeliveryFilesMover
from cg.services.deliver_files.rsync.service import DeliveryRsyncService
from cg.services.deliver_files.tag_fetcher.abstract import FetchDeliveryFileTagsService
from cg.services.deliver_files.tag_fetcher.bam_service import BamDeliveryTagsFetcher
from cg.services.deliver_files.tag_fetcher.sample_and_case_service import (
    SampleAndCaseDeliveryTagsFetcher,
)
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.store.models import Case
from cg.store.store import Store


class DeliveryServiceFactory:
    """Class to build the delivery services based on workflow and delivery type."""

    def __init__(
        self,
        store: Store,
        hk_api: HousekeeperAPI,
        rsync_service: DeliveryRsyncService,
        tb_service: TrailblazerAPI,
        analysis_service: AnalysisService,
    ):
        self.store = store
        self.hk_api = hk_api
        self.rsync_service = rsync_service
        self.tb_service = tb_service
        self.analysis_service = analysis_service

    @staticmethod
    def _sanitise_delivery_type(delivery_type: DataDelivery) -> DataDelivery:
        """Sanitise the delivery type."""
        if delivery_type in [DataDelivery.FASTQ_QC, DataDelivery.FASTQ_SCOUT]:
            return DataDelivery.FASTQ
        if delivery_type in [DataDelivery.ANALYSIS_SCOUT]:
            return DataDelivery.ANALYSIS_FILES
        if delivery_type in [
            DataDelivery.FASTQ_ANALYSIS_SCOUT,
            DataDelivery.FASTQ_QC_ANALYSIS,
        ]:
            return DataDelivery.FASTQ_ANALYSIS
        return delivery_type

    @staticmethod
    def _validate_delivery_type(delivery_type: DataDelivery):
        """Check if the delivery type is supported. Raises DeliveryTypeNotSupported error."""
        if delivery_type in [
            DataDelivery.FASTQ,
            DataDelivery.ANALYSIS_FILES,
            DataDelivery.FASTQ_ANALYSIS,
            DataDelivery.BAM,
        ]:
            return
        raise DeliveryTypeNotSupported(
            f"Delivery type {delivery_type} is not supported. Supported delivery types are"
            f" {DataDelivery.FASTQ}, {DataDelivery.ANALYSIS_FILES},"
            f" {DataDelivery.FASTQ_ANALYSIS}, {DataDelivery.BAM}."
        )

    @staticmethod
    def _get_file_tag_fetcher(delivery_type: DataDelivery) -> FetchDeliveryFileTagsService:
        """Get the file tag fetcher based on the delivery type."""
        service_map: dict[DataDelivery, Type[FetchDeliveryFileTagsService]] = {
            DataDelivery.FASTQ: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.ANALYSIS_FILES: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.FASTQ_ANALYSIS: SampleAndCaseDeliveryTagsFetcher,
            DataDelivery.BAM: BamDeliveryTagsFetcher,
        }
        return service_map[delivery_type]()

    def _get_file_fetcher(self, delivery_type: DataDelivery) -> FetchDeliveryFilesService:
        """Get the file fetcher based on the delivery type."""
        service_map: dict[DataDelivery, Type[FetchDeliveryFilesService]] = {
            DataDelivery.FASTQ: RawDataDeliveryFileFetcher,
            DataDelivery.ANALYSIS_FILES: AnalysisDeliveryFileFetcher,
            DataDelivery.FASTQ_ANALYSIS: RawDataAndAnalysisDeliveryFileFetcher,
            DataDelivery.BAM: RawDataDeliveryFileFetcher,
        }
        file_tag_fetcher: FetchDeliveryFileTagsService = self._get_file_tag_fetcher(delivery_type)
        return service_map[delivery_type](
            status_db=self.store,
            hk_api=self.hk_api,
            tags_fetcher=file_tag_fetcher,
        )

    def _convert_workflow(self, case: Case) -> Workflow:
        """Converts a workflow with the introduction of the microbial-fastq delivery type an
        unsupported combination of delivery type and workflow setup is required. This function
        makes sure that a raw data workflow with microbial fastq delivery type is treated as a
        microsalt workflow so that the microbial-fastq sample files can be concatenated."""
        tag: str = case.samples[0].application_version.application.tag
        microbial_tags: list[str] = [
            application.tag
            for application in self.store.get_active_applications_by_prep_category(
                prep_category=SeqLibraryPrepCategory.MICROBIAL
            )
        ]
        if case.data_analysis == Workflow.RAW_DATA and tag in microbial_tags:
            return Workflow.MICROSALT
        return case.data_analysis

    def _get_sample_file_formatter(
        self,
        case: Case,
    ) -> SampleFileFormatter | SampleFileConcatenationFormatter:
        """Get the file formatter service based on the workflow."""
        converted_workflow: Workflow = self._convert_workflow(case)
        if converted_workflow in [Workflow.MICROSALT]:
            return SampleFileConcatenationFormatter(
                file_manager=FileManagingService(),
                file_formatter=SampleFileNameFormatter(),
                concatenation_service=FastqConcatenationService(),
            )
        return SampleFileFormatter(
            file_manager=FileManagingService(), file_name_formatter=SampleFileNameFormatter()
        )

    def build_delivery_service(
        self, case: Case, delivery_type: DataDelivery | None = None
    ) -> DeliverFilesService:
        """Build a delivery service based on a case."""
        delivery_type: DataDelivery = self._sanitise_delivery_type(
            delivery_type if delivery_type else case.data_delivery
        )
        self._validate_delivery_type(delivery_type)
        file_fetcher: FetchDeliveryFilesService = self._get_file_fetcher(delivery_type)
        sample_file_formatter: SampleFileFormatter | SampleFileConcatenationFormatter = (
            self._get_sample_file_formatter(case)
        )
        file_formatter: DeliveryFileFormattingService = DeliveryFileFormatter(
            case_file_formatter=CaseFileFormatter(), sample_file_formatter=sample_file_formatter
        )
        return DeliverFilesService(
            delivery_file_manager_service=file_fetcher,
            move_file_service=DeliveryFilesMover(),
            file_filter=SampleFileFilter(),
            file_formatter_service=file_formatter,
            status_db=self.store,
            rsync_service=self.rsync_service,
            tb_service=self.tb_service,
            analysis_service=self.analysis_service,
        )



"""Delivery API path fixtures."""

from pathlib import Path

import pytest

from cg.constants import FileExtensions


@pytest.fixture
def delivery_fastq_file(tmp_path: Path, sample_id: str) -> Path:
    file = Path(tmp_path, f"{sample_id}_R1_001{FileExtensions.FASTQ_GZ}")
    file.touch()
    return file


@pytest.fixture
def delivery_bam_file(tmp_path: Path, sample_id: str) -> Path:
    file = Path(tmp_path, f"{sample_id}_R1_001{FileExtensions.BAM}")
    file.touch()
    return file


@pytest.fixture
def delivery_another_fastq_file(tmp_path: Path, another_sample_id: str) -> Path:
    file = Path(tmp_path, f"{another_sample_id}_R1_001{FileExtensions.FASTQ_GZ}")
    file.touch()
    return file


@pytest.fixture
def delivery_another_bam_file(tmp_path: Path, another_sample_id: str) -> Path:
    file = Path(tmp_path, f"{another_sample_id}_R1_001{FileExtensions.BAM}")
    file.touch()
    return file


@pytest.fixture
def delivery_spring_file(tmp_path: Path, sample_id: str) -> Path:
    file = Path(tmp_path, f"{sample_id}_S1_001{FileExtensions.SPRING}")
    file.touch()
    return file


@pytest.fixture
def delivery_another_spring_file(tmp_path: Path, another_sample_id: str) -> Path:
    file = Path(tmp_path, f"{another_sample_id}_S1_001{FileExtensions.SPRING}")
    file.touch()
    return file


@pytest.fixture
def delivery_report_file(tmp_path: Path, case_id: str) -> Path:
    file = Path(tmp_path, f"{case_id}_delivery-report{FileExtensions.HTML}")
    file.touch()
    return file


@pytest.fixture
def delivery_cram_file(tmp_path: Path, sample_id: str) -> Path:
    file = Path(tmp_path, f"{sample_id}{FileExtensions.CRAM}")
    file.touch()
    return file


@pytest.fixture
def delivery_another_cram_file(tmp_path: Path, another_sample_id: str) -> Path:
    file = Path(tmp_path, f"{another_sample_id}{FileExtensions.CRAM}")
    file.touch()
    return file


@pytest.fixture
def delivery_ticket_dir_path(tmp_path: Path, ticket_id: str) -> Path:
    return Path(tmp_path, ticket_id)
